/*
 * Output driver encapulation for trackedit.
 *
 * Adding a driver is a fairly simple matter.  The driver must provide an
 * init() function which initialises the type-specific portion of the
 * Driver::dev union in the Driver structure, and sets the open, write,
 * close and free function pointers (and optionally the configure and
 * get_buffer_size function pointers) in the Driver structure to point to
 * the appropriate functions within the driver module.  A driver identifier
 * should be added to driver.h; driver_new() then only has to have the
 * appropriate type-specific init() call added to give a fully functioning
 * driver.
 * 
 * The init function and Driver function pointers expect to point to a
 * function taking a single `Device *' parameter, except for write which
 * also expects a `char *' buffer and an `int' size parameter in addition to
 * the `Driver *' parameter.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "drivers.h"

#define DEFAULT_SAMPLE_RATE          44100
#define DEFAULT_SAMPLE_SIZE          16
#define DEFAULT_N_CHANNELS           2
#define DEFAULT_DRIVER_BUFFER_SIZE   2048

/* ======================================================================== */
/* Prototypes for externally provided driver routines */

/* From drv_ossdsp.c */
int drv_ossdsp_init(Driver *d);
int drv_file_init(Driver *d);

/* ======================================================================== */

Driver *driver_new(int type) {
/*
 * Creates a new driver of the specified type.
 *
 */
Driver *res = malloc(sizeof(Driver));
int i = 1;
  if (res==NULL)
    return(NULL);
  res->type = type;
  res->sample_rate = DEFAULT_SAMPLE_RATE;
  res->n_channels = DEFAULT_N_CHANNELS;
  res->sample_size = DEFAULT_SAMPLE_SIZE;
  res->active = 0;
  res->configure = NULL;
  res->get_buffer_size = NULL;
  res->open = NULL; res->write = NULL;
  res->close = NULL; res->free = NULL;
  switch (type) {
    case drvOSSDSP:
      i = drv_ossdsp_init(res);
      break;
    case drvFILE:
      i = drv_file_init(res);
      break;
    default:
      fprintf(stderr,"driver_new(): Unknown driver type %d\n",type);
      i = 0;
  }
  if (i<1) {
    free(res);
    res = NULL;
  }
  return(res);
}
/* ======================================================================== */

int driver_configure(Driver *d, int sample_rate, int n_channels, 
                     int sample_size) {
/*
 * Requests that a driver reconfigure itself to suit new sample settings.
 *
 */
  if (d->active) {
    fprintf(stderr,"driver_configure(): called when active\n");
    return 0;
  }
  d->sample_rate = sample_rate;
  d->n_channels = n_channels;
  d->sample_size = sample_size;
  if (d->configure!=NULL)
    return d->configure(d,sample_rate,n_channels,sample_size);
  else
    return 1;
}
/* ======================================================================== */

int driver_open(Driver *d) {
/*
 * Opens the given driver ready to output data.  Dispatches control to the
 * appropriate driver.  Returns 1 on success, 0 on failure.
 *
 */
  if (d->active) {
    fprintf(stderr,"driver_open(): called when already active\n");
    return(0);
  }
  d->total_written = 0;
  d->active = (d->open(d) == 1);
  return(d->active);
}
/* ======================================================================== */

int driver_get_buffer_size(Driver *d) {
/*
 * Returns the `ideal' buffer size to use with the driver.  Should be
 * called after driver_open() for most reliable results, since some drivers
 * (OSS for example) only know the optimum buffer size after opening.
 *
 */
int ret = DEFAULT_DRIVER_BUFFER_SIZE;
  if (d->get_buffer_size!=NULL)
    ret = d->get_buffer_size(d);
  return(ret);
}
/* ======================================================================== */

int driver_write(Driver *d, char *buf, int size) {
/*
 * Writes the data in the buffer to the specified device.  The number of bytes
 * actually written is returned.  The cumulative frame count is maintained
 * in d->total_written.
 *
 */
int ret;
  if (!d->active) {
    fprintf(stderr,"driver_write(): called when driver not active\n");
    return(-1);
  }
  ret = d->write(d,buf,size);
  if (ret)
    d->total_written += ret/((d->sample_size>>3)*d->n_channels);
  return ret;
}
/* ======================================================================== */

int driver_close(Driver *d) {
/*
 * Closes the driver - data can't be sent again until driver_open() is called.
 * Returns 1 on success, 0 on failure.
 *
 */
  if (!d->active) {
    fprintf(stderr,"driver_close(): called when driver not active\n");
    return(0);
  }
  d->active = (d->close(d) != 1);
  return(!d->active);
}
/* ======================================================================== */

int driver_free(Driver *d) {
/*
 * Disposes of the given driver.
 *
 */
  if (d==NULL)
    return(1);
  if (d->active)
    d->close(d);
  d->free(d);
  free(d);
  return(1);
}
/* ======================================================================== */
