/*
 * The text-interface module for trackedit.
 *
 */
#include <math.h>

#include <sys/time.h>        /* For select() */
#include <sys/types.h>
#include <unistd.h>

#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <readline/readline.h>
#include <readline/history.h>

#include <errno.h>

#include "drivers.h"
#include "trackedit.h"

enum if_text_commands {
  ifNO_COMMAND = -2,
  ifUNKNOWN_COMMAND = -1,
  ifEXIT = 0,
  ifFBACKWARD,
  ifFADE,
  ifFFORWARD,
  ifGOTOMARK,
  ifHELP,
  ifLOAD,
  ifPLAY,
  ifPOS,
  ifQUIT,
  ifSETMARK,
  ifSHOWMARKS,
  ifSTATUS,
  ifSTOP,
};

enum if_play_flags {
  pfSTOP = 0,
  pfPLAY,
  pfFFORWARD,
  pfFBACKWARD,
};

char *playmodestr[] = {"stopped","playing","fast playing forwards", 
                    "fast playing backwards"};

typedef int (*Commandhandler)(Trackeditor *, char *);

typedef struct Commandinfo {
  char *command;
  int id;
  char *help;
  Commandhandler handler;
} Commandinfo;

#define MAIN_PROMPT        "trackedit: "

/* ======================================================================== */
#define N_MARKS 10         /* Number of marks useable */

Commandinfo commands[];
int run;                   /* =1 until the interface should shut down */
int playmode;              /* !=0 if currently playing audio */
float ff_speed;
float fadein_time, fadeout_time;
Trackeditor *te;           /* A track editor associated with the interface */
int marks[N_MARKS];        /* Frame number associated with each mark */

Driver *monitor;           /* Monitor device */

/* ======================================================================== */

void clear_marks() {
int i;
  for (i=0;i<N_MARKS;i++)
    marks[i] = -1;
}
/* ======================================================================== */

char *skip_whitespace(char *s) {
char *c = s;
  while (c!=NULL && (*c==' ' || *c=='\t'))
    c++;
  return c;
}
/* ======================================================================== */

int parse_command(char *cmd, char **options) {
/*
 * Identifies a command in cmd and returns its identifier.  options is set
 * to point to the command parameters if there are any.
 *
 */
char *tindex, *cindex;
int f1=-1, f2=-1, c = 0;
  /* Trim leading white space */
  while (*cmd==' ' || *cmd=='\t')
    cmd++;
  if (*cmd==0)
    return(ifNO_COMMAND);

  /* Identify the command */
  while (commands[c].command!=NULL && (f1==-1 || f2==-1)) {
    tindex = commands[c].command;
    cindex = cmd;
    while (*tindex==*cindex && *tindex!='\0' && *cindex!='\0') {
      tindex++;
      cindex++;
    }
    if (cindex!=cmd && (*cindex=='\0' || *cindex==' ')) {
      if (f1<0)
        f1 = c;
      else
        f2 = c;
    }
    c++;
  }
  /* Find parameters if specified */
  if (f1>=0 && f2<0) {
    char *args = cmd;
    while (*args!='\0' && *args!=' ')
      args++;
    *options = args + (*args==' ');
    if (**options=='\0')
      *options=NULL;
    return(commands[f1].id);
  } else {
    return(ifUNKNOWN_COMMAND);
  }
}
/* ======================================================================== */

int parse_frame_position(Trackeditor *te, char *s, int *pos, int origin) {
/*
 * Parses a position from the given string and returns it as a number of
 * frames.  Either a time (hh:mm:ss.xxxx) or sample number may be specified. 
 * Returns the number of characters consumed by the position specification
 * on success, and 0 on failure. `+' or `-' can preceed the position, in
 * which case it is interpreted as an offset from the current position (if
 * origin==-1) or the given origin.
 *
 */
char *c = s;
char *token, *comma = NULL;
char lastsep = 0;
int is_negative = 0;
int is_positive = 0;
int f[4];
int n, i, ret = 1;
int n_dp = 0;
float fpos = 0.0;
  *pos = 0;
 
  /* Skip leading whitespace */
  c = skip_whitespace(c);

  /* Deal immediately with a marker specification */
  if (*c=='m') {
    c++;
    if (sscanf(c,"%d",&n)==0)
      return 0;
    sscanf(c,"%*d%n",&ret);
    if (n<1 || n>N_MARKS)
      return 0;
    if (marks[n-1]==-1)
      return 0;
    *pos = marks[n-1];
    return ret+1;
  }
  if ((is_negative = *c=='-') == 1)
    c++;
  else
  if ((is_positive = *c=='+') == 1)
    c++;

  /* Work out whether we have a frame number or time.  We need to discount */
  /* anything after a comma or whitespace since we're only interested in   */
  /* first time of the first argument. */
  i = 0;
  comma=strchr(c,',');
  token=strchr(c,' ');
  if (comma==NULL || (token!=NULL && token<comma))
    comma=token;
  token=strchr(c,'\t');
  if (comma==NULL || (token!=NULL && token<comma))
    comma=token;
  if (comma!=NULL) {
    while (comma!=c && *comma!=':' && *comma!='.')
      comma--;
    /* If separators have been found, we conclude we have a time */
    if (comma!=c)
      i=1;
  } else
    i = strchr(c,':')!=NULL || strchr(c,'.')!=NULL;
  if (i!=0) {
    /* A time's been specified */
    i = 0;
    while (ret && *c!=0 && *c!='\t' && *c!=' ' && i<4) {
      /* sscanf() man page says not to rely on the effect of %n on the */
      /* return value, so call it twice: once to get the number of     */
      /* characters read, and one to find out whether %d was processed */
      /* without error. */
      sscanf(c,"%d%n",&f[i],&n);
      ret = sscanf(c,"%d",&f[i]) == 1;
      i++;
      if (c!=s)
        lastsep = *(c-1);
      /* The the forth separator must be `.', not a `:' */
      if (c!=s && *(c-1)==':' && i==4) {
        ret = 0;
        break;
      }
      c+=n;
      /* Point past the separator if present */
      if (*c=='.' || *c==':')
        c++;
      /* Only one field allowed after a decimal point */
      if (c!=s && lastsep=='.') {
        n_dp = n; /* Store the number of characters after the decimal point */
        break;
      }
      /* If the previous character was not a time separator,we've come */
      /* to the end of the time. */
      if (*(c-1)!=':' && *(c-1)!='.')
        break;
    }
    /* Skip any trailing whitespace */
    c = skip_whitespace(c);

    /* Calculate the number of characters consumed by time specification */
    if (ret)
      ret = (c-s);
    if (ret) {
      n = 0;
      while (n<i-(lastsep=='.')) {
        fpos = (fpos)*60 + f[n++];
      }
      if (lastsep=='.' && f[n]!=0) {
        fpos += f[n] / powf(10,n_dp);
      }
      *pos = trackeditor_get_frames_of_time(te,fpos);
    }
  } else {
    /* Parse a simple frame number */
    ret = sscanf(c,"%d",&f[0]) == 1;
    if (ret) {
      *pos = f[0];
      sscanf(c,"%*d%n",&ret);
      /* Include the +/- character in the returned character count */
      if (is_positive || is_negative)
        ret++;
    }
  }
  if (is_negative)
    *pos *= -1;
  if (ret && (is_positive || is_negative)) {
    if (origin==-1)
      *pos += trackeditor_get_frame_position(te);
    else
      *pos += origin;
  }
  if (*pos < 0)
    *pos = 0;
  if (*pos >= te->frame_count)
    *pos = te->frame_count-1;
  return ret;
}

/* ======================================================================== */

char *parse_tofile_arg(char **args) {
/*
 * Parses a `file=' argument in the given argument list and returns a
 * pointer to a newly allocated copy of the actual filename.  If there is no
 * `file=' parameter at the start of `args', NULL is returned.  args is
 * updated to point to the delimiting character at the end of the tofile
 * parameter.
 *
 */
char *ap = skip_whitespace(*args);
char *file = NULL;
int n,i;

  if (*args==NULL)
    return NULL;
  /* Extract an output filename */
  if (strstr(ap,"file=")==ap) {
    ap+=5;
    i=n=0;
    while (ap[n]!=' ' && ap[n]!='\t' && ap[n]!=0)
      n++;
    file = malloc((n+1)*sizeof(char));
    while (i<n) {
      file[i++] = ap[i];
    }
    file[n]=0;
    *args += n;
  }
  return file;
}
/* ======================================================================== */

char *position_string(Trackeditor *te, int frame) {
/*
 * Formulates a string describing the given frame position in the 
 * trackeditor.
 *
 */
static char buf[40];
float time = trackeditor_get_time_of_frames(te,frame);

  sprintf(buf,"%02d:%02d:%02d.%03d (sample %d)",
      (int)(time/3600), ((int)(time)/60) % 60,
      (int)(time) % 60, (int)((time-(int)time)*1000),frame);
  return(buf);
}
/* ======================================================================== */

void show_fadetimes(int which) {
/*
 * Displays the current fade times.  If `which' is <0, both are shown. 
 * Otherwise only the fadein (which==0) or fadeout (which==1) is shown.
 *
 */
  if (which<=0) {
    printf("Play fade-in time: ");
    if (fadein_time<=0)
      printf("<not in use>\n");
    else
      printf("%g seconds\n",fadein_time);
  }
  if (which<0 || which==1) {
    printf("Play fade-out time: ");
    if (fadeout_time<=0)
      printf("<not in use>\n");
    else
      printf("%g seconds\n",fadeout_time);
  }
}
/* ======================================================================== */

int setup_fastplay(Trackeditor *te, char *args, int forward) {
/*
 * Sets up the common parts of a fast play request (either forwards or
 * backwards) at a speed indicated by args.  If no speed is supplied, it
 * defaults to 5.
 *
 */
char *modestr[] = {"backward","forward"};
float start_time;
int start_frame;
float speed = ff_speed;

  forward = (forward?1:0);
  if (te->input==NULL) {
    fprintf(stderr,"Cannot fast play %s - no input defined\n",modestr[forward]);
    return 0;
  }
  if (playmode!=pfSTOP) {
    fprintf(stderr,"Currently %s - use `stop' to stop\n",playmodestr[playmode]);
    return 0;
  }
  start_frame = trackeditor_get_frame_position(te);
  if (args!=NULL) {
    if (sscanf(args,"%f",&speed)!=1) {
      fprintf(stderr,"Invalid speed specified: %s\n",args);
      return 0;
    }
    ff_speed = speed;
  }    
  start_time = trackeditor_get_time_of_frames(te,start_frame);
  printf("Fast playing %ss at %gx speed from %s\n",modestr[forward],ff_speed,
    position_string(te,start_frame));
  trackeditor_set_frame_position(te,start_frame);

  /* Open driver ready to accept output */
  driver_open(te->output);

  /* Activate playing */
  if (forward)
    playmode = pfFFORWARD;
  else
    playmode = pfFBACKWARD;
  return 1;
}
/* ======================================================================== */

int file_exists(char *filename) {
/*
 * Returns 1 if the specified file exists.
 *
 */
struct stat buf;
  return stat(filename,&buf)==0;
}
/* ======================================================================== */

char *readline_info(char *thisprompt) {
/*
 * Uses readline() to get some info for the program with the supplied prompt.
 * Restores the program's main prompt before returning.
 *
 */
char *ret;
  rl_save_prompt();
  ret = readline(thisprompt);
  rl_restore_prompt();
  return ret;
}
/* ======================================================================== */

int trackeditor_setpos(Trackeditor *te, char *args) {
/*
 * Sets the trackeditor's position to that specified in the arguments.  Used
 * by if_play() and if_pos().
 *
 */
int start_frame;
int i;
  if (playmode!=pfSTOP) {
    fprintf(stderr,"Currently %s - use `stop' to stop\n",playmodestr[playmode]);
    return 0;
  }
  if (args==NULL)
    start_frame = trackeditor_get_frame_position(te);
  else
  if ((i=parse_frame_position(te,args,&start_frame,-1))<1 || 
      start_frame>=te->frame_count || start_frame<0) {
    fprintf(stderr,"Invalid position specified\n");
    return 0;
  }
  trackeditor_set_frame_position(te,start_frame);
  return i;
}
/* ======================================================================== */

int if_fbackward(Trackeditor *te, char *args) {
/*
 * As for `if_fforward', but move through the file backwards from the
 * current position.
 *
 */
  return setup_fastplay(te,args,0);
}
/* ======================================================================== */

int if_fade(Trackeditor *te, char *args) {
/*
 * Sets fade-in and fade-out times.
 *
 */
char *ap = args;
int which;
int clear = 0;
float newtime = -1;

  /* Default to dealing with both fade times */
  which = -1;
  /* Deal with supplied arguments */
  if (args!=NULL) {
    ap = skip_whitespace(ap);
    /* Deduce which fade time to deal with */
    if (strstr(ap,"in")==ap) {
      ap+=2;
      which = 0;
    } else
    if (strstr(ap,"out")==ap) {
      which = 1;
      ap+=3;
    }
    ap = skip_whitespace(ap);
    if (*ap!=0) {
      if (!(clear = strstr(ap,"off")==ap)) {
        if (sscanf(ap,"%f",&newtime)!=1 || newtime<0) {
          fprintf(stderr,"Bad parameters: use `help fade' for details\n");
          return 0;
        }
        if (newtime==0) {
          newtime = -1;
          clear = 1;
        }
      }
    }
  }
  /* Carry out the requested task */
  if (!clear && newtime==-1) {
    show_fadetimes(which);
  } else {
    if (which<=0)
      fadein_time = newtime;
    if (which<0 || which==1) 
      fadeout_time = newtime; 
  }
  return 1;
}
/* ======================================================================== */

int if_fforward(Trackeditor *te, char *args) {
/*
 * Fast forwards from the current position at a speed indicated by args.  If
 * no speed is supplied, it defaults to 5.
 *
 * This is really just another interface to the `play' command, allowing the
 * speed to be set rather than the start position.
 *
 */
  return setup_fastplay(te,args,1);
}
/* ======================================================================== */

int if_gotomark(Trackeditor *te, char *args) {
/* 
 * Positions the track editor at the given mark.
 *
 */
int mark;
  if (playmode!=pfSTOP) {
    fprintf(stderr,"Currently %s - use `stop' to stop\n",playmodestr[playmode]);
    return 0;
  }
  if (te->input==NULL) {
    fprintf(stderr,"No input loaded into track editor: cannot set position\n");
    return 0;
  }
  if (args==NULL || sscanf(args,"%d",&mark)!=1 || mark<1 || mark>N_MARKS) {
    fprintf(stderr,"Bad or missing mark number (must be between 1 and %d\n",N_MARKS);
    return 0;
  }
  if (marks[mark-1]==-1) {
    fprintf(stderr,"Mark %d not set\n",mark);
    return 0;
  }
  trackeditor_set_frame_position(te,marks[mark-1]);
  printf("Position now %s\n",position_string(te,marks[mark-1]));
  return 1;
}
/* ======================================================================== */

int if_help(Trackeditor *te, char *args) {
/*
 * Displays a command help message.  If a command is named as an option,
 * more detailed help on that command is given.
 */
char *c,*scratch;
int i,cmd;
Commandinfo *command = commands;
char buf[256];
  if (args!=NULL)
    cmd = parse_command(args,&scratch);
  else
    cmd = ifUNKNOWN_COMMAND;

  if (cmd==ifUNKNOWN_COMMAND || cmd==ifNO_COMMAND) {
    fprintf(stderr,"Available commands (unique truncations also accepted):\n");
    while (command->command!=NULL) {
      fprintf(stderr,"  %s",command->command);
      if (command->help==NULL || command->help[0]=='\0') {
        command++;
        fprintf(stderr,"\n");
        continue;
      }
      if ((scratch = strchr(command->help,'\t')) == NULL)
        scratch = command->help;
      else
        scratch++;
      c = strchr(scratch,'.');
      if (c!=NULL) {
        buf[0]='\0';
        i = 1;
        snprintf(buf,c-scratch+4,": %s",scratch);
        fprintf(stderr,"%s\n",buf);
      } else
        fprintf(stderr,"\n");
      command++;
    }
  } else {
    /* Help on specific command */
    command = &commands[cmd];
    fprintf(stderr,command->command);
    if (command->help==NULL || command->help[0]=='\0') {
      fprintf(stderr,": no extended help available\n");
      return(1);
    }
    c = strchr(command->help,'\t');
    if (c==NULL) {
      fprintf(stderr,"\n");
      c = command->help;
    } else {
      snprintf(buf,c-command->help+3," %s",command->help);
      fprintf(stderr,"%s\n",buf);
      c++;
    }
    while (*c!='\0') {
      strcpy(buf,"  ");
      i=2;
      while (*c!='\n' && *c!='\0') {
        buf[i++] = *(c++);
      }
      buf[i++]='\0';
      fprintf(stderr,"%s\n",buf);
      if (*c!='\0')
        c++;
    }
  }
  return(1);
}
/* ======================================================================== */

int if_load(Trackeditor *te, char *args) {
/*
 * Loads the supplied file into the given trackeditor as input.
 *
 */
char *fname = args;
int ret = 1, rls = 0, i;
  if (playmode!=pfSTOP) {
    fprintf(stderr,"Currently %s - use `stop' before load\n",playmodestr[playmode]);
    return 0;
  }
  if (fname == NULL) {
    rls = (fname = readline_info("Filename to load: ")) != NULL;
  }
  if (fname==NULL)
    return 0;
  if (strlen(fname) != 0) {
    /* Trim trailing whitespace */
    while (fname[i=strlen(fname)-1]==' ' || fname[i]=='\t')
      fname[i]=0;
    if (!file_exists(fname) || trackeditor_set_input_file(te,fname)==0) {
      fprintf(stderr,"File `%s' could not be opened\n",fname);
      ret = 0;
    } else
      clear_marks();
  }
  if (rls)
    free(fname);
  return ret;
}
/* ======================================================================== */

int if_play(Trackeditor *te, char *args) {
/*
 * Commence playing from the current (or a specified) position.
 *
 */
Driver *out;
char *file = NULL;
char *ap = args;
int start_frame, end_frame;
int i=0,j=0;
  if (te->input==NULL) {
    fprintf(stderr,"Cannot play - no input defined\n");
    return 0;
  }
  file = parse_tofile_arg(&ap);
  if (ap!=NULL) {
    if ((i=trackeditor_setpos(te,ap))<1)
      return 0;
    ap += i;
  }
  /* Read an `end position' if specified */
  if (ap!=NULL && *ap==',') {
    if ((j=parse_frame_position(te,ap+1,&end_frame,-1))<1) {
      fprintf(stderr,"Bad end time/position specification\n");
      return 0;
    }
    ap += (j+1);
  } else
    end_frame = -1;
  if (file==NULL)
    file = parse_tofile_arg(&ap);
  if (file!=NULL && file[0]==0) {
    fprintf(stderr,"Missing filename\n");
    return 0;
  }

  start_frame = trackeditor_get_frame_position(te);
  if (end_frame!=-1 && end_frame<start_frame) {
    fprintf(stderr,"End time/position less than start position - ignored\n");
    end_frame = -1;
  }
  trackeditor_set_play_options(te,start_frame,end_frame,fadein_time,fadeout_time);
  printf("Playing from %s",position_string(te,start_frame));
  if (end_frame>0) {
    printf(" to %s\n",position_string(te,end_frame));
  } else
    printf("\n");

  /* Deal with the case where we're not outputting to a specified file */
  if (file==NULL) {
    /* Open driver ready to accept output */
    driver_open(te->output);

    /* Activate playing */
    playmode = pfPLAY;
    return 1;
  }
  /* Output is to a file, so open a new device, associate it with the */
  /* trackeditor's output, play into it and then restore everything   */
  /* before exitting. */
  out = driver_new(drvFILE);
  /* The driver will free() this string on cleanup */
  out->dev.file.filename=file;
  trackeditor_set_output(te,out);
  if (driver_open(out)<1) {
    fprintf(stderr,"Could not open output file %s\n",file);
    trackeditor_set_output(te,monitor);
    driver_free(out);
    return 0;
  }
  while (((i=trackeditor_get_frame_position(te))<te->frame_count) &&
         (te->end_frame==-1 || i<te->end_frame)) {
    printf("\rWriting: current position is %s%n",position_string(te,i),&j);
    fflush(stdout);
    /* Write 0.5 seconds between updates */
    trackeditor_play(te,-1,0.5);
  }
  /* Print final completion message and blank anything left behind */
  /* by the last "Writing: ..." message. */
  printf("\rWriting to file %s complete.%n",file,&i);
  while (j>i++)
    printf(" ");
  printf("\n");
  driver_close(out);
  trackeditor_set_output(te,monitor);
  driver_free(out);
  return 1;
}
/* ======================================================================== */

int if_pos(Trackeditor *te, char *args) {
/*
 * Commence sets the trackeditor's position to that specified in the
 * arguments.
 *
 */
int start_frame;
  if (playmode!=pfSTOP) {
    fprintf(stderr,"Currently %s - use `stop' to stop\n",playmodestr[playmode]);
    return 0;
  }
  if (trackeditor_setpos(te,args)!=1)
    return 0;
  start_frame = trackeditor_get_frame_position(te);
  printf("Position now %s\n",position_string(te,start_frame));
  return 1;
}
/* ======================================================================== */

int if_quit(Trackeditor *te, char *args) {
/*
 * Cause the text interface to exit.
 *
 */
  run = 0;
  return 1;
}
/* ======================================================================== */

int if_setmark(Trackeditor *te, char *args) {
/* 
 * Sets a mark.
 *
 */
int mark_num = -1;
int frame = -1;
int n, origin;
  if (args!=NULL) {
    sscanf(args,"%d%n",&mark_num,&n);
  }
  if (mark_num<1 || mark_num>N_MARKS) {
    fprintf(stderr,"Bad or missing mark number (must be betweeen 1 and %d)\n",N_MARKS);
    return 0;
  }
  if (te==NULL) {
    fprintf(stderr,"Can't set marks without an input for the trackeditor\n");
    return 0;
  }
  while (*(args+n)==' ' || *(args+n)=='\t')
    n++;
  origin = marks[mark_num-1];
  if (origin==-1)
    origin=0;
  /* If no frame specified, default to the current position */
  if (*(args+n)==0)
    frame = trackeditor_get_frame_position(te);
  else
  if (strstr(args+n,"clear")==args+n)
    frame = -1;
  else
  if (parse_frame_position(te,args+n,&frame,origin)<1) {
    fprintf(stderr,"Invalid position specified\n");
    return 0;
  }
  marks[mark_num-1] = frame;
  if (frame==-1) 
    printf("Mark %2d cleared\n",mark_num);
  else
    printf("Mark %2d set at %s\n",mark_num,position_string(te,frame));
  return 1;
}
/* ======================================================================== */

int if_showmarks(Trackeditor *te, char *args) {
/*
 * Displays information about a mark, or all marks if none are specified.
 *
 */
int num = -1;
int i;
  if (args!=NULL && sscanf(args,"%d",&num)!=1) {
    fprintf(stderr,"Bad mark number\n");
    return 0;
  }
  if (num!=-1 && (num<1 || num>N_MARKS)) {
    fprintf(stderr,"Mark number must be between 1 and %d\n",N_MARKS);
    return 0;
  }
  for (i=(num==-1?0:num-1);i<(num==-1?N_MARKS:num);i++) {
    if (marks[i]!=-1)
      printf("Mark %2d at %s\n",i+1,position_string(te,marks[i]));
    else
      printf("Mark %2d is not set\n",i+1);
  }
  return 1;
}
/* ======================================================================== */

int if_status(Trackeditor *te, char *args) {
/*
 * Displays general program status information.
 *
 */
int curr_frame = 0;

  if (te->input != NULL) {
    curr_frame = trackeditor_get_frame_position(te);
    printf("Input filename: %s\n",te->input_filename);
    printf("Current position: %s\n",position_string(te,curr_frame));
    printf("Last input frame: %s\n",position_string(te,te->frame_count));
    printf("Sample rate = %d Hz, number of channels = %d, sample size = %d bits\n",
      te->sample_rate, te->channel_count, te->sample_size);
  } else {
    printf("No input currently open\n");
  }

  printf("Fast play speed: %g times normal\n",ff_speed);
  show_fadetimes(-1);

  if (playmode==0)
    printf("Output idle\n");
  else {
    printf("Currently %s to output\n",playmodestr[playmode]); 
  }
  return 1;
}
/* ======================================================================== */

int if_stop(Trackeditor *te, char *args) {
/*
 * Stops a current play session.
 *
 */
  if (playmode==pfSTOP) {
    fprintf(stderr,"Nothing is playing - cannot stop\n");
    return 0;
  }
  trackeditor_stopplay(te);
  playmode = pfSTOP;
  printf("Stopped at %s\n",position_string(te,trackeditor_get_frame_position(te)));
  return 1;
}
/* ======================================================================== */

int dispatch_command(char *cmd, Trackeditor *te) {
/*
 * Identifies a command and calls the relevant handler.  Cmd may contain
 * options after the command itself.
 *
 */
int i;
char *param = NULL;

  /* Identify the command and any options present */
  i = parse_command(cmd,&param);
  if (i==ifNO_COMMAND)
    return(1);
  if (i==ifUNKNOWN_COMMAND) {
    fprintf(stderr,"Unknown command: `%s'\n",cmd);
    return(0);
  }
  /* Dispatch the command to the relevant handler */
  return commands[i].handler(te,param);
}
/* ======================================================================== */

/* Table of available commands */
Commandinfo commands[] = {
  { command: "exit",    id: ifQUIT,
    help: "Exit the program.", 
    handler: if_quit,
  },
  { command: "fbackward", id: ifFBACKWARD,
    help: "[<speed>]\t"
          "Fast play backwards at <speed> times normal speed.  Sound is played into\n"
          "the active output.  If omitted <speed> defaults to the last speed used, or\n"
          "5 if there has been no previous fforward/fbackward command.",
    handler: if_fbackward,
  },
  { command: "fade", id: ifFADE,
    help: "[\"in\"|\"out\"] [<time>|\"off\"]\t"
          "Set fade-in/fade-out times.  If neither \"in\" nor \"out\" are specified,\n"
          "changes are made to both.  If the last parameter is omitted, the current\n"
          "times are reported; if \"off\" is specifed, the applicable fade is cancelled.\n"
          "<time> is the fade time in seconds.\n",
    handler: if_fade,
  },
  { command: "fforward",id: ifFFORWARD,
    help: "[<speed>]\t"
          "Fast play forwards at <speed> times normal speed.  Sound is played into\n"
          "the active output.  If omitted <speed> defaults to the last speed used, or\n"
          "5 if there has been no previous fforward/fbackward command.",
    handler: if_fforward,
  },
  { command: "gotomark", id: ifGOTOMARK,
    help: "<mark number>\t"
          "Moves the trackeditor to mark <mark number>.",
    handler: if_gotomark,
  },
  { command: "help",    id: ifHELP,
    help: "[<command>]\tDisplay general help, or help on a specific command.",
    handler: if_help,
  },
  { command: "load",	id: ifLOAD,
    help: "[<filename>]\tLoads a new input filename.  If not supplied, a prompt is issued.",
    handler: if_load,
  },
  { command: "play",	id: ifPLAY,
    help: "[[\"+\"|\"-\"]<position>][,<endpos>] [\"file=\"<filename>]\t"
          "Play input from the current position, or from the position indicated.\n"
          "<position> may be a frame number, a mark number if preceeded by `m' or a\n"
          "time of the form hh:mm:ss.ddd, hh:mm:ss, mm:ss.ddd, mm:ss or ss.ddd where\n"
          "hh=hour, mm=minute, ss=seconds and ddd=decimal portion of seconds.  If `+'\n"
          "or `-' preceeds <position>, <position> is interpreted as an offset from the\n"
          "current position (except if a marker is specified).  <endpos>, if given, is\n"
          "the playback end position in the same format used for <position>.  If\n"
          "<endpos> is omitted, playback ends at the end-of-input.\n"
          "When the \"file=...\" argument is supplied, data is written to the specified\n"
          "file (<filename>) rather than the normal output device.",
    handler: if_play,
  },
  { command: "position",id: ifPOS,
    help: "[<position>]\t"
          "Position the track editor at the given position.  <position> can be specified in\n"
          "the same way as for the `play' command.\n",
    handler: if_pos,
  },
  { command: "quit",    id: ifQUIT,
    help: "Exit the program.", 
    handler: if_quit,
  },
  { command: "setmark", id: ifSETMARK,
    help: "<mark number> [<position>|\"clear\"]\t"
          "sets mark number <mark number>.  If <position> is given it is used to set\n"
          "the mark; if \"clear\" is givem the mark is cleared; otherwise the track\n"
          "editor's current position is used.  <position> can be specified as for the\n"
          "`play' command.  If `+' or `-' preceeds <position>, <position is interpreted\n"
          "as an offset from the current mark's position (or 0 if the mark is unset).",
    handler: if_setmark,
  },
  { command: "showmarks", id: ifSHOWMARKS,
    help: "[<mark number>]\t"
          "Shows the position of a mark.  All marks are shown if <mark number> is\n"
          "omitted.",
    handler: if_showmarks,
  },
  { command: "status",	id: ifSTATUS,
    help: "Displays program status information.",
    handler: if_status,
  },
  { command: "stop",	id: ifSTOP,
    help: "Stop current play operation.",
    handler: if_stop,
  },
};
/* ======================================================================== */

void readline_handler(char *line) {
/*
 * Called by readline when there is a complete line to handle.
 *
 */
  if (line==NULL || line[0]==0)
    return;
  add_history(line);
  dispatch_command(line,te);

  /* If we are about to exit, reinstall the callback handler without a   */
  /* prompt.  This stops readline putting a prompt on the line following */
  /* the quitting command. */
  if (!run)
    rl_callback_handler_install("", readline_handler);

  if (line!=NULL)
    free(line);
}
/* ======================================================================== */

int if_text(Options *options) {
/*
 * The text interface entry-point.
 *
 */
struct timeval tv;
fd_set fds;
int i;
  using_history();
  stifle_history(40);
  run = 1;
  ff_speed = 5;
  fadein_time = fadeout_time = -1;
  clear_marks();

  /* Set up a TrackEditor for the interface.  Load a file if requested. */
  te = trackeditor_new();
  if (options->init_inputfile!=NULL) {
    if (!file_exists(options->init_inputfile) ||
        trackeditor_set_input_file(te,options->init_inputfile)!=1) {
      fprintf(stderr,"Input file %s could not be opened\n",options->init_inputfile);
    }
  }

  /* The default output is the monitor (currently an OSS PCM device) */
  monitor = driver_new(drvOSSDSP);
  monitor->dev.ossdsp.device=strdup("/dev/dsp");
  trackeditor_set_output(te,monitor);

  rl_callback_handler_install(MAIN_PROMPT, readline_handler);
  while (run) {
    FD_ZERO(&fds);
    FD_SET(fileno(stdin),&fds);
    tv.tv_sec = 0;
    tv.tv_usec = 10*1000;  /* 10ms */
    i = select(fileno(stdin)+1,&fds,NULL,NULL,&tv);
    if (i<0 && i!=EINTR)
      break;
    if (i>0 && FD_ISSET(fileno(stdin), &fds))
      rl_callback_read_char();

    if (playmode==pfSTOP) {
      /* Sleep for 10ms so CPU load doesn't go through the roof */
      usleep(10*1000);
    } else {
      switch (playmode) {
        case pfPLAY:
          /* Play 20ms of audio per cycle */
          trackeditor_play(te,-1,0.02);
          break;
        case pfFFORWARD: {
          int pos = trackeditor_get_frame_position(te);
          trackeditor_play(te,-1,0.05);
          trackeditor_set_frame_position(te,pos+
            trackeditor_get_frames_of_time(te,ff_speed*0.05));
          break;
        }
        case pfFBACKWARD: {
          int pos = trackeditor_get_frame_position(te);
          trackeditor_play(te,-1,0.05);
          trackeditor_set_frame_position(te,pos-
            trackeditor_get_frames_of_time(te,ff_speed*0.05));
          break;
        }
      }
      if (te->end_frame>0 && 
          trackeditor_get_frame_position(te)>=te->end_frame) {
        printf("\rEnd position reached: ");
        if_stop(te,NULL);
        rl_refresh_line(0,0);
      } else
      if (trackeditor_get_frame_position(te)>=te->frame_count-1) {
        printf("\rEnd of input: ");
        if_stop(te,NULL);
        rl_refresh_line(0,0);
      } else
      if (trackeditor_get_frame_position(te)==0) {
        printf("\rAt start of input: ");
        if_stop(te,NULL);
        rl_refresh_line(0,0);
      }
    }
  }
  clear_history();
  driver_free(monitor);
  if (te!=NULL)
    trackeditor_free(te);
  return(1);
}
/* ======================================================================== */
